from __future__ import annotations
import asyncio
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
import typing as t
import warnings

import asyncpg
from asyncpg.connection import Connection
from asyncpg.cursor import Cursor
from asyncpg.pool import Pool
from asgiref.sync import async_to_sync

from piccolo.engine.base import Batch, Engine
from piccolo.query.base import Query
from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync
from piccolo.utils.warnings import colored_warning


@dataclass
class AsyncBatch(Batch):

    connection: Connection
    query: Query
    batch_size: int

    # Set internally
    _transaction = None
    _cursor: t.Optional[Cursor] = None

    @property
    def cursor(self) -> Cursor:
        if not self._cursor:
            raise ValueError("_cursor not set")
        return self._cursor

    async def next(self) -> t.List[t.Dict]:
        data = await self.cursor.fetch(self.batch_size)
        return self.query._process_results(data)

    def __aiter__(self):
        return self

    async def __anext__(self):
        response = await self.next()
        if response == []:
            raise StopAsyncIteration()
        return response

    async def __aenter__(self):
        self._transaction = self.connection.transaction()
        await self._transaction.start()
        template, template_args = self.query.querystring.compile_string()

        self._cursor = await self.connection.cursor(template, *template_args)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        print("waiting to exit")
        await self._transaction.commit()
        await self.connection.close()


###############################################################################


class Transaction:
    """
    Usage:

    transaction = engine.Transaction()
    transaction.add(Foo.create())
    transaction.run_sync()
    """

    __slots__ = ("engine", "queries")

    def __init__(self, engine: PostgresEngine):
        self.engine = engine
        self.queries: t.List[Query] = []

    def add(self, *query: Query):
        self.queries += list(query)

    async def _run_queries(self, connection):
        async with connection.transaction():
            for query in self.queries:
                _query, args = query.querystring.compile_string(
                    engine_type=self.engine.engine_type
                )
                await connection.execute(_query, *args)

        self.queries = []

    async def _run_in_pool(self):
        pool = await self.engine.get_pool()
        connection = await pool.acquire()

        try:
            await self._run_queries(connection)
        except Exception:
            pass
        finally:
            await pool.release(connection)

    async def _run(self):
        connection = await asyncpg.connect(**self.engine.config)
        await self._run_queries(connection)

    async def run(self, in_pool=True):
        if in_pool and self.engine.pool:
            await self._run_in_pool()
        else:
            await self._run()

    def run_sync(self):
        return run_sync(self._run())


###############################################################################
class PostgresEngine(Engine):

    __slots__ = ("config", "pool")

    engine_type = "postgres"
    min_version_number = 9.6

    def __init__(self, config: t.Dict[str, t.Any]) -> None:
        self.config = config
        self.pool: t.Optional[Pool] = None
        super().__init__()

    def get_version(self) -> float:
        """
        Returns the version of Postgres being run.
        """
        loop = asyncio.new_event_loop()
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(
                loop.run_until_complete, self.run("SHOW server_version")
            )

        response: t.Array[t.Dict] = future.result()
        server_version = response[0]["server_version"]
        major, minor, _ = server_version.split(".")
        version = float(f"{major}.{minor}")
        return version

    ###########################################################################

    async def start_connnection_pool(self, **kwargs) -> None:
        if self.pool:
            warnings.warn(
                "A pool already exists - close it first if you want to create a new pool."
            )
        else:
            config = dict(self.config)
            config.update(**kwargs)
            self.pool = await asyncpg.create_pool(**config)

    async def close_connnection_pool(self) -> None:
        if self.pool:
            await self.pool.close()
            self.pool = None
        else:
            warnings.warn("No pool is running.")

    ###########################################################################

    async def get_connection(self) -> Connection:
        return await asyncpg.connect(**self.config)

    ###########################################################################

    async def batch(self, query: Query, batch_size=100) -> AsyncBatch:
        connection = await self.get_connection()
        return AsyncBatch(
            connection=connection, query=query, batch_size=batch_size
        )

    ###########################################################################

    async def run_in_pool(self, query: str, args: t.List[t.Any] = []):
        if not self.pool:
            raise ValueError("A pool isn't currently running.")

        async with self.pool.acquire() as connection:
            response = await connection.fetch(query, *args)

        return response

    async def run(self, query: str, args: t.List[t.Any] = []):
        connection = await self.get_connection()
        results = await connection.fetch(query, *args)
        await connection.close()
        return results

    async def run_querystring(
        self, querystring: QueryString, in_pool: bool = True
    ):
        query, args = querystring.compile_string(engine_type=self.engine_type)
        if in_pool and self.pool:
            return await self.run_in_pool(query, args)
        else:
            return await self.run(query, args)

    def transaction(self) -> Transaction:
        return Transaction(engine=self)



import asyncio

from unittest import TestCase

from ..base import DBTestCase, postgres_only
from ..example_project.tables import Manager


@postgres_only
class TestPool(DBTestCase):
    async def create_pool(self):
        await Manager._meta.db.start_connnection_pool()
        await Manager._meta.db.close_connnection_pool()

    async def make_queries(self):
        await Manager._meta.db.start_connnection_pool()

        await Manager(name="Bob").save().run()
        response = await Manager.select().run()
        self.assertTrue("Bob" in [i["name"] for i in response])

        await Manager._meta.db.close_connnection_pool()

    def test_creation(self):
        """
        Make sure a connection pool can be created.
        """
        asyncio.run(self.create_pool())

    def test_queries(self):
        """
        Make several queries using a connection pool.
        """
        asyncio.run(self.make_queries())



from __future__ import annotations
import typing as t

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Create(Query):
    """
    Creates a database table.
    """

    __slots__: t.Tuple = tuple()

    @property
    def querystring(self) -> QueryString:
        base = f"CREATE TABLE {self.table._meta.tablename}"
        columns = ", ".join(["{}" for i in self.table._meta.columns])
        query = f"{base} ({columns})"
        return QueryString(
            query, *[i.querystring for i in self.table._meta.columns]
        )

    def __str__(self) -> str:
        return self.querystring.__str__()



from __future__ import annotations
import dataclasses
import itertools
import typing as t

from piccolo.columns.base import Column
from piccolo.query.base import Query
from piccolo.querystring import QueryString


@dataclasses.dataclass
class AlterStatement:
    __slots__ = ("column",)

    column: Column

    def querystring(self) -> QueryString:
        raise NotImplementedError()

    def __str__(self) -> str:
        return self.querystring.__str__()


@dataclasses.dataclass
class Rename(AlterStatement):
    __slots__ = ("new_name",)

    new_name: str

    @property
    def querystring(self) -> QueryString:
        return QueryString(
            f"RENAME COLUMN {self.column._meta.name} TO {self.new_name}"
        )


@dataclasses.dataclass
class Drop(AlterStatement):
    @property
    def querystring(self) -> QueryString:
        return QueryString(f"DROP {self.column._meta.name}")


@dataclasses.dataclass
class Add(AlterStatement):
    __slots__ = ("name",)

    name: str

    @property
    def querystring(self) -> QueryString:
        self.column._meta.name = self.name
        return QueryString("ADD {}", self.column.querystring)


@dataclasses.dataclass
class Unique(AlterStatement):
    __slots__ = ("boolean",)

    boolean: bool

    @property
    def querystring(self) -> QueryString:
        if self.boolean:
            return QueryString(f"ADD UNIQUE ({self.column._meta.name})")
        else:
            tablename = self.column._meta.table._meta.tablename
            column_name = self.column._meta.name
            key = f"{tablename}_{column_name}_key"
            return QueryString(f'DROP CONSTRAINT "{key}"')


@dataclasses.dataclass
class Null(AlterStatement):
    __slots__ = ("boolean",)

    boolean: bool

    @property
    def querystring(self) -> QueryString:
        if self.boolean:
            return QueryString(f"{self.column._meta.name} DROP NOT NULL")
        else:
            return QueryString(f"{self.column._meta.name} SET NOT NULL")


class Alter(Query):
    __slots__ = ("_add", "_drop", "_rename", "_unique", "_null")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._add: t.List[Add] = []
        self._drop: t.List[Drop] = []
        self._rename: t.List[Rename] = []
        self._unique: t.List[Unique] = []
        self._null: t.List[Null] = []

    def add_column(self, name: str, column: Column) -> Alter:
        """
        Band.alter().add_column(‘members’, Integer())
        """
        self._add.append(Add(column, name))
        return self

    def drop_column(self, column: Column) -> Alter:
        """
        Band.alter().drop_column('popularity')
        """
        self._drop.append(Drop(column))
        return self

    def rename_column(self, column: Column, new_name: str) -> Alter:
        """
        Band.alter().rename_column(Band.popularity, ‘rating’)
        """
        self._rename.append(Rename(column, new_name))
        return self

    def set_null(self, column: Column, boolean: bool = True) -> Alter:
        """
        Band.alter().set_null(True)
        """
        self._null.append(Null(column, boolean))
        return self

    def set_unique(self, column: Column, boolean: bool = True) -> Alter:
        """
        Band.alter().set_unique(True)
        """
        self._unique.append(Unique(column, boolean))
        return self

    @property
    def querystring(self) -> QueryString:
        query = f"ALTER TABLE {self.table._meta.tablename}"

        alterations = [
            i.querystring
            for i in itertools.chain(
                self._add, self._rename, self._drop, self._unique, self._null
            )
        ]

        for a in alterations:
            query += " {}"

        return QueryString(query, *alterations)

    def __str__(self) -> str:
        return self.querystring.__str__()



from unittest import TestCase

from ..example_project.tables import Band, Manager
from ..base import postgres_only


@postgres_only
class TestTransaction(TestCase):
    def test_error(self):
        """
        Make sure queries in a transaction aren't committed if a query fails.
        """
        transaction = Band._meta.db.transaction()
        transaction.add(
            Manager.create(),
            Band.create(),
            Band.raw("MALFORMED QUERY ... SHOULD ERROR"),
        )
        try:
            transaction.run_sync()
        except Exception:
            pass
        self.assertTrue(not Band.table_exists().run_sync())
        self.assertTrue(not Manager.table_exists().run_sync())

    def test_succeeds(self):
        transaction = Band._meta.db.transaction()
        transaction.add(Manager.create(), Band.create())
        transaction.run_sync()

        self.assertTrue(Band.table_exists().run_sync())
        self.assertTrue(Manager.table_exists().run_sync())

        transaction.add(Band.drop(), Manager.drop())
        transaction.run_sync()



from __future__ import annotations
import typing as t

from piccolo.query.base import Query
from piccolo.query.mixins import AddDelegate
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:
    from piccolo.table import Table


class Insert(Query):
    __slots__ = ("add_delegate",)

    def setup_delegates(self):
        self.add_delegate = AddDelegate()

    def add(self, *instances: Table) -> Insert:
        self.add_delegate.add(*instances, table_class=self.table)
        return self

    def run_callback(self, results):
        """
        Assign the ids of the created rows to the model instances.
        """
        for index, row in enumerate(results):
            self.add_delegate._add[index].id = row["id"]

    @property
    def sqlite_querystring(self) -> QueryString:
        base = f"INSERT INTO {self.table._meta.tablename}"
        columns = ",".join([i._meta.name for i in self.table._meta.columns])
        values = ",".join(["{}" for i in self.add_delegate._add])
        query = f"{base} ({columns}) VALUES {values}"
        return QueryString(
            query,
            *[i.querystring for i in self.add_delegate._add],
            query_type="insert",
        )

    @property
    def postgres_querystring(self) -> QueryString:
        base = f"INSERT INTO {self.table._meta.tablename}"
        columns = ",".join([i._meta.name for i in self.table._meta.columns])
        values = ",".join(["{}" for i in self.add_delegate._add])
        query = f"{base} ({columns}) VALUES {values} RETURNING id"
        return QueryString(
            query,
            *[i.querystring for i in self.add_delegate._add],
            query_type="insert",
        )



from piccolo.columns import Integer

from ..base import DBTestCase
from ..example_project.tables import Band, Manager


class TestRename(DBTestCase):
    def test_rename(self):
        self.insert_row()

        rename_query = Band.alter().rename_column(Band.popularity, "rating")
        rename_query.run_sync()

        select_query = Band.raw("SELECT * FROM band")
        response = select_query.run_sync()

        column_names = response[0].keys()
        self.assertTrue(
            ("rating" in column_names) and ("popularity" not in column_names)
        )


class TestDrop(DBTestCase):
    def test_drop(self):
        self.insert_row()

        Band.alter().drop_column(Band.popularity).run_sync()

        response = Band.raw("SELECT * FROM band").run_sync()

        column_names = response[0].keys()
        self.assertTrue("popularity" not in column_names)


class TestAdd(DBTestCase):
    def test_add(self):
        """
        This needs a lot more work. Need to set values for existing rows.

        Just write the test for now ...
        """
        self.insert_row()

        add_query = Band.alter().add_column("weight", Integer())
        add_query.run_sync()

        response = Band.raw("SELECT * FROM band").run_sync()

        column_names = response[0].keys()
        self.assertTrue("weight" in column_names)

        self.assertEqual(response[0]["weight"], None)


class TestUnique(DBTestCase):
    def test_unique(self):
        unique_query = Manager.alter().set_unique(Manager.name, True)
        unique_query.run_sync()

        Manager(name="Bob").save().run_sync()

        # Make sure non-unique names work:
        Manager(name="Sally").save().run_sync()

        # Check there's a unique row error ...
        with self.assertRaises(Exception):
            Manager(name="Bob").save().run_sync()

        response = Manager.select().run_sync()
        self.assertTrue(len(response) == 2)

        # Now remove the constraint, and add a row.
        not_unique_query = Manager.alter().set_unique(Manager.name, False)
        not_unique_query.run_sync()
        Manager(name="Bob").save().run_sync()

        response = Manager.select().run_sync()
        self.assertTrue(len(response), 2)


class TestNull(DBTestCase):
    def test_null(self):
        pass


class TestMultiple(DBTestCase):
    pass



from __future__ import annotations
import copy
from dataclasses import dataclass, field
import typing as t

from piccolo.engine import Engine, engine_finder
from piccolo.columns import Column, PrimaryKey, ForeignKey
from piccolo.query import (
    Alter,
    Create,
    Count,
    Delete,
    Drop,
    Exists,
    Insert,
    Objects,
    Raw,
    Select,
    TableExists,
    Update,
)
from piccolo.querystring import QueryString, Unquoted
from piccolo.utils import _camel_to_snake


@dataclass
class TableMeta:
    """
    This is used to store info about the table.
    """

    tablename: str = ""
    columns: t.List[Column] = field(default_factory=list)
    non_default_columns: t.List[Column] = field(default_factory=list)
    db: t.Optional[Engine] = engine_finder()

    def get_column_by_name(self, name: str) -> Column:
        column = [i for i in self.columns if i._meta.name == name]
        if len(column) != 1:
            raise ValueError(f"No matching column found with name == {name}")
        return column[0]


class TableMetaclass(type):
    def __str__(cls):
        """
        Returns a basic string representation of the table and its columns.

        Used by the playground.
        """
        spacer = "\n    "
        columns = []
        for col in cls._meta.columns:
            if isinstance(col, ForeignKey):
                columns.append(
                    f"{col._meta.name} = ForeignKey({col._foreign_key_meta.references.__name__})"
                )
            else:
                columns.append(
                    f"{col._meta.name} = {col.__class__.__name__}()"
                )
        columns_string = spacer.join(columns)
        return f"class {cls.__name__}(Table):\n" f"    {columns_string}\n"


class Table(metaclass=TableMetaclass):

    # These are just placeholder values, so type inference isn't confused - the
    # actual values are set in __init_subclass__.
    _meta = TableMeta()
    id = PrimaryKey()

    def __init_subclass__(
        cls, tablename: t.Optional[str] = None, db: t.Optional[Engine] = None
    ):
        """
        Automatically populate the _meta, which includes the tablename, and
        columns.
        """
        cls.id = PrimaryKey()

        tablename = tablename if tablename else _camel_to_snake(cls.__name__)

        attribute_names = [i for i in dir(cls) if not i.startswith("_")]

        columns: t.List[Column] = []
        non_default_columns: t.List[Column] = []

        for attribute_name in attribute_names:
            attribute = getattr(cls, attribute_name)
            if isinstance(attribute, Column):
                column = attribute
                columns.append(column)

                if not isinstance(column, PrimaryKey):
                    non_default_columns.append(column)

                column._meta._name = attribute_name
                # Mypy wrongly thinks cls is a Table instance:
                column._meta._table = cls  # type: ignore

        db = db if db else engine_finder()

        cls._meta = TableMeta(
            tablename=tablename, columns=columns, non_default_columns=[], db=db
        )

    def __init__(self, **kwargs):
        """
        Assigns any default column values to the class.
        """
        for column in self._meta.columns:
            value = kwargs.pop(column._meta.name, None)
            if not value:
                if column.default:
                    # Can't use inspect - can't tell that datetime.datetime.now
                    # is a callable.
                    is_callable = hasattr(column.default, "__call__")
                    value = column.default() if is_callable else column.default
                else:
                    if not column._meta.null:
                        raise ValueError(
                            f"{column._meta.name} wasn't provided"
                        )
            self[column._meta.name] = value

        unrecognized = kwargs.keys()
        if unrecognized:
            raise ValueError(f"Unrecognized columns - {unrecognized}")

    ###########################################################################

    def save(self) -> t.Union[Insert, Update]:
        """
        A proxy to an insert or update query.
        """
        if not hasattr(self, "id"):
            raise ValueError("No id value found")

        cls = self.__class__

        if type(self.id) == int:
            # pre-existing row
            kwargs: t.Dict[Column, t.Any] = {
                i: getattr(self, i._meta.name, None)
                for i in cls._meta.columns
                if i._meta.name != "id"
            }
            return cls.update().values(kwargs).where(cls.id == self.id)
        else:
            return cls.insert().add(self)

    def remove(self) -> Delete:
        """
        A proxy to a delete query.
        """
        _id = self.id

        if type(_id) != int:
            raise ValueError("Can only delete pre-existing rows with an id.")

        self.id = None  # type: ignore

        return self.__class__.delete().where(self.__class__.id == _id)

    def get_related(self, foreign_key: ForeignKey) -> Objects:
        """
        Used to fetch a Table instance, for the target of a foreign key.

        band = await Band.objects().first().run()
        manager = await band.get_related(Band.name).run()
        >>> print(manager.name)
        'Guido'

        It can only follow foreign keys one level currently.
        i.e. Band.manager, but not Band.manager.x.y.z

        """
        cls = self.__class__

        if isinstance(foreign_key, ForeignKey):
            column_name = foreign_key._meta.name

            references: t.Type[
                Table
            ] = foreign_key._foreign_key_meta.references

            return (
                references.objects()
                .where(
                    references._meta.get_column_by_name("id")
                    == getattr(self, column_name)
                )
                .first()
            )
        else:
            raise ValueError(f"{column_name} isn't a ForeignKey")

    def __setitem__(self, key: str, value: t.Any):
        setattr(self, key, value)

    def __getitem__(self, key: str):
        return getattr(self, key)

    ###########################################################################

    @property
    def querystring(self) -> QueryString:
        """
        Used when inserting rows.
        """
        args_dict = {
            col._meta.name: self[col._meta.name] for col in self._meta.columns
        }

        is_unquoted = lambda arg: type(arg) == Unquoted

        # Strip out any args which are unquoted.
        # TODO Not the cleanest place to have it (would rather have it handled
        # in the QueryString bundle logic) - might need refactoring.
        filtered_args = [i for i in args_dict.values() if not is_unquoted(i)]

        # If unquoted, dump it straight into the query.
        query = ",".join(
            [
                args_dict[column._meta.name].value
                if is_unquoted(args_dict[column._meta.name])
                else "{}"
                for column in self._meta.columns
            ]
        )
        return QueryString(f"({query})", *filtered_args)

    def __str__(self) -> str:
        return self.querystring.__str__()

    ###########################################################################
    # Classmethods

    @classmethod
    def ref(cls, column_name: str) -> Column:
        """
        Used to get a copy of a column in a reference table.

        Example: manager.name
        """
        local_column_name, reference_column_name = column_name.split(".")

        local_column = cls._meta.get_column_by_name(local_column_name)

        if not isinstance(local_column, ForeignKey):
            raise ValueError(f"{local_column_name} isn't a ForeignKey")

        reference_column = local_column.references._meta.get_column_by_name(
            reference_column_name
        )

        _reference_column = copy.deepcopy(reference_column)
        _reference_column.name = f"{local_column_name}.{reference_column_name}"
        return _reference_column

    @classmethod
    def insert(cls, *rows: "Table") -> Insert:
        """
        await Band.insert(
            Band(name="Pythonistas", popularity=500, manager=1)
        ).run()
        """
        query = Insert(table=cls)
        if rows:
            query.add(*rows)
        return query

    @classmethod
    def raw(cls, sql: str, *args: t.Any) -> Raw:
        """
        Execute raw SQL queries on the underlying engine - use with caution!

        await Band.raw('select * from band')

        Or passing in parameters:

        await Band.raw("select * from band where name = {}", 'Pythonistas')
        """
        return Raw(table=cls, base=QueryString(sql, *args))

    @classmethod
    def select(cls) -> Select:
        """
        Get data in the form of a list of dictionaries, with each dictionary
        representing a row.

        await Band.select().columns(Band.name).run()
        """
        return Select(table=cls)

    @classmethod
    def delete(cls) -> Delete:
        """
        await Band.delete().where(Band.name == 'Pythonistas').run()
        """
        return Delete(table=cls)

    @classmethod
    def create(cls) -> Create:
        """
        Create table, along with all columns.

        await Band.create().run()
        """
        return Create(table=cls)

    @classmethod
    def create_without_columns(cls) -> Raw:
        """
        Create the table, but with no columns (useful for migrations).

        await Band.create_without_columns().run()
        """
        return Raw(
            table=cls,
            base=QueryString(f'CREATE TABLE "{cls._meta.tablename}"()'),
        )

    @classmethod
    def drop(cls) -> Drop:
        """
        Drops the table.

        await Band.drop().run()
        """
        return Drop(table=cls)

    @classmethod
    def alter(cls) -> Alter:
        """
        Used to modify existing tables and columns.

        await Band.alter().rename_column(Band.popularity, 'rating')
        """
        return Alter(table=cls)

    @classmethod
    def objects(cls) -> Objects:
        """
        Returns a list of table instances (each representing a row), which you
        can modify and then call 'save' on, or can delete by calling 'remove'.

        pythonistas = await Band.objects().where(
            Band.name == 'Pythonistas'
        ).first().run()

        pythonistas.name = 'Pythonistas Reborn'

        await pythonistas.save().run()

        # Or to remove it from the database:
        await pythonistas.remove()
        """
        return Objects(table=cls)

    @classmethod
    def count(cls) -> Count:
        """
        Count the number of matching rows.

        await Band.count().where(Band.popularity > 1000).run()
        """
        return Count(table=cls)

    @classmethod
    def exists(cls) -> Exists:
        """
        Use it to check if a row exists, not if the table exists.

        await Band.exists().where(Band.name == 'Pythonistas').run()
        """
        return Exists(table=cls)

    @classmethod
    def table_exists(cls) -> TableExists:
        """
        Check if the table exists in the database.

        await Band.table_exists().run()
        """
        return TableExists(table=cls)

    @classmethod
    def update(cls) -> Update:
        """
        Update rows.

        await Band.update().values(
            {Band.name: "Spamalot"}
        ).where(Band.name=="Pythonistas")
        """
        return Update(table=cls)



from __future__ import annotations

from piccolo.custom_types import Combinable
from piccolo.query.base import Query
from piccolo.query.mixins import WhereDelegate
from piccolo.querystring import QueryString


class Delete(Query):

    __slots__ = ("where_delegate",)

    def setup_delegates(self):
        self.where_delegate = WhereDelegate()

    def where(self, where: Combinable) -> Delete:
        self.where_delegate.where(where)
        return self

    @property
    def querystring(self) -> QueryString:
        query = f"DELETE FROM {self.table._meta.tablename}"
        if self.where_delegate._where:
            query += " WHERE {}"
            return QueryString(query, self.where_delegate._where.querystring)
        else:
            return QueryString(query)

    def __str__(self) -> str:
        query = f"DELETE FROM {self.table._meta.tablename}"
        if self.where_delegate._where:
            query += f" WHERE {self.where_delegate._where.__str__()}"
        return query



from __future__ import annotations

from piccolo.custom_types import Combinable
from piccolo.query.base import Query
from piccolo.query.mixins import WhereDelegate
from piccolo.querystring import QueryString

from .select import Select


class Exists(Query):
    __slots__ = ("where_delegate",)

    def setup_delegates(self):
        self.where_delegate = WhereDelegate()

    def where(self, where: Combinable) -> Exists:
        self.where_delegate.where(where)
        return self

    def response_handler(self, response) -> bool:
        return response[0]["exists"]

    @property
    def querystring(self) -> QueryString:
        select = Select(
            self.table,
            QueryString(f"SELECT * FROM {self.table._meta.tablename}"),
        )
        select.where_delegate._where = self.where_delegate._where
        return QueryString('SELECT EXISTS({}) AS "exists"', select.querystring)

    def __str__(self) -> str:
        return self.querystring.__str__()



from __future__ import annotations
import typing as t

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class TableExists(Query):

    __slots__: t.Tuple = tuple()

    def response_handler(self, response):
        return bool(response[0]["exists"])

    @property
    def sqlite_querystring(self) -> QueryString:
        return QueryString(
            "SELECT EXISTS(SELECT * FROM sqlite_master WHERE "
            f"name = '{self.table._meta.tablename}') AS 'exists'"
        )

    @property
    def postgres_querystring(self) -> QueryString:
        return QueryString(
            "SELECT EXISTS(SELECT * FROM information_schema.tables WHERE "
            f"table_name = '{self.table._meta.tablename}')"
        )

    def __str__(self) -> str:
        return self.querystring.__str__()



from __future__ import annotations
import typing as t

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Raw(Query):
    __slots__: t.Tuple = tuple()

    @property
    def querystring(self) -> QueryString:
        return self.base



from abc import abstractmethod, ABCMeta

from piccolo.utils.warnings import colored_warning


class Batch:
    pass


class Engine(metaclass=ABCMeta):
    def __init__(self):
        self.check_version()

    @property
    @abstractmethod
    def engine_type(self) -> str:
        pass

    @property
    @abstractmethod
    def min_version_number(self) -> float:
        pass

    @abstractmethod
    def get_version(self) -> float:
        pass

    def check_version(self):
        """
        Warn if the database version isn't supported.
        """
        version_number = self.get_version()
        print(f"Running version {version_number}")
        if version_number < self.min_version_number:
            message = (
                f"This version of {self.engine_type} isn't supported "
                f"(< {self.min_version_number}) - some features might not be "
                "available. For instructions on installing databases, see the "
                "Piccolo docs."
            )
            colored_warning(message, stacklevel=3)



import warnings

import colorama


colorama.init()


def colored_warning(message: str, stacklevel: int = 3):
    colored_message = colorama.Fore.RED + message + colorama.Fore.RESET
    warnings.warn(colored_message, stacklevel=stacklevel)



from __future__ import annotations
import itertools
import typing as t

import ujson as json

from piccolo.querystring import QueryString
from piccolo.utils.sync import run_sync

if t.TYPE_CHECKING:
    from piccolo.table import Table  # noqa


class Query(object):
    def __init__(
        self,
        table: t.Type[Table],
        base: QueryString = QueryString(""),
        *args,
        **kwargs,
    ) -> None:
        self.base = base
        self.table = table
        self.setup_delegates()
        super().__init__()

    def setup_delegates(self):
        pass

    @property
    def engine_type(self) -> str:
        engine = self.table._meta.db
        if engine:
            return engine.engine_type
        else:
            raise ValueError("Engine isn't defined.")

    def _process_results(self, results):
        if results:
            keys = results[0].keys()
            keys = [i.replace("$", ".") for i in keys]
            raw = [dict(zip(keys, i.values())) for i in results]
        else:
            raw = []

        if hasattr(self, "run_callback"):
            self.run_callback(raw)

        # TODO - I have multiple ways of modifying the final output
        # response_handlers, and output ...
        # Might try and merge them.
        raw = self.response_handler(raw)

        output = getattr(self, "output_delegate", None)

        if output:
            if output._output.as_objects:
                # When using .first() we get a single row, not a list
                # of rows.
                if type(raw) is list:
                    raw = [self.table(**columns) for columns in raw]
                else:
                    raw = self.table(**raw)
            elif type(raw) is list:
                if output._output.as_list:
                    if len(raw[0].keys()) != 1:
                        raise ValueError(
                            "Each row returned more than one value"
                        )
                    else:
                        raw = list(itertools.chain(*[j.values() for j in raw]))
                if output._output.as_json:
                    raw = json.dumps(raw)

        return raw

    async def run(self, in_pool=True):
        engine = getattr(self.table._meta, "db", None)
        if not engine:
            raise ValueError(
                f"Table {self.table._meta.tablename} has no db defined in _meta"
            )

        results = await engine.run_querystring(
            self.querystring, in_pool=in_pool
        )

        return self._process_results(results)

    def run_sync(self, *args, **kwargs):
        """
        A convenience method for running the coroutine synchronously.
        """
        return run_sync(self.run(*args, **kwargs, in_pool=False))

    def response_handler(self, response):
        """
        Subclasses can override this to modify the raw response returned by
        the database driver.
        """
        return response

    ###########################################################################

    @property
    def sqlite_querystring(self) -> QueryString:
        raise NotImplementedError

    @property
    def postgres_querystring(self) -> QueryString:
        raise NotImplementedError

    @property
    def default_querystring(self) -> QueryString:
        raise NotImplementedError

    @property
    def querystring(self) -> QueryString:
        """
        Calls the correct underlying method, depending on the current engine.
        """
        engine_type = self.engine_type
        if engine_type == "postgres":
            try:
                return self.postgres_querystring
            except NotImplementedError:
                return self.default_querystring
        elif engine_type == "sqlite":
            try:
                return self.sqlite_querystring
            except NotImplementedError:
                return self.default_querystring
        else:
            raise Exception(
                f"No querystring found for the {engine_type} engine."
            )

    ###########################################################################

    def __str__(self) -> str:
        return self.querystring.__str__()



from dataclasses import dataclass
import sqlite3
import typing as t

import aiosqlite
from aiosqlite import Cursor, Connection

from piccolo.engine.base import Batch, Engine
from piccolo.query.base import Query
from piccolo.querystring import QueryString
from piccolo.utils.warnings import colored_warning


@dataclass
class AsyncBatch(Batch):

    connection: Connection
    query: Query
    batch_size: int

    # Set internally
    _cursor: t.Optional[Cursor] = None

    @property
    def cursor(self) -> Cursor:
        if not self._cursor:
            raise ValueError("_cursor not set")
        return self._cursor

    async def next(self) -> t.List[t.Dict]:
        data = await self.cursor.fetchmany(self.batch_size)
        return self.query._process_results(data)

    def __aiter__(self):
        return self

    async def __anext__(self):
        response = await self.next()
        if response == []:
            raise StopAsyncIteration()
        return response

    async def __aenter__(self):
        template, template_args = self.query.querystring.compile_string()

        self._cursor = await self.connection.execute(template, *template_args)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        print("waiting to exit")
        await self._cursor.close()
        await self.connection.close()


###############################################################################


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class SQLiteEngine(Engine):

    __slots__ = ("path",)

    engine_type = "sqlite"
    min_version_number = 3.25

    def __init__(self, path: str = "piccolo.sqlite") -> None:
        self.path = path
        super().__init__()

    def get_version(self) -> float:
        """
        Warn if the version of SQLite isn't supported.
        """
        major, minor, _ = sqlite3.sqlite_version_info
        return float(f"{major}.{minor}")

    ###########################################################################

    async def batch(self, query: Query, batch_size=100) -> AsyncBatch:
        connection = await self.get_connection()
        return AsyncBatch(
            connection=connection, query=query, batch_size=batch_size
        )

    ###########################################################################

    async def get_connection(self) -> Connection:
        connection = await aiosqlite.connect(
            self.path, detect_types=sqlite3.PARSE_DECLTYPES
        )
        connection.row_factory = dict_factory
        return connection

    ###########################################################################

    async def run(
        self, query: str, args: t.List[t.Any] = [], query_type: str = "generic"
    ):
        async with aiosqlite.connect(
            self.path, detect_types=sqlite3.PARSE_DECLTYPES
        ) as connection:

            connection.row_factory = dict_factory
            async with connection.execute(query, args) as cursor:
                cursor.row_factory = dict_factory
                await connection.commit()
                response = await cursor.fetchall()
                # print(query)
                # print(args)
                # print(response)

                # print(query_type)
                if query_type == "insert":
                    return [{"id": cursor.lastrowid}]
                else:
                    return response

    async def run_querystring(
        self, querystring: QueryString, in_pool: bool = False
    ):
        return await self.run(
            *querystring.compile_string(engine_type=self.engine_type),
            query_type=querystring.query_type,
        )

    def transaction(self):
        raise NotImplementedError



from __future__ import annotations
import typing as t

from piccolo.query.base import Query
from piccolo.querystring import QueryString


class Drop(Query):

    __slots__: t.Tuple = tuple()

    @property
    def querystring(self) -> QueryString:
        return QueryString(f'DROP TABLE "{self.table._meta.tablename}"')

    def __str__(self) -> str:
        return self.querystring.__str__()



from __future__ import annotations

from piccolo.custom_types import Combinable
from piccolo.query.base import Query
from piccolo.query.mixins import WhereDelegate
from piccolo.querystring import QueryString

from .select import Select


class Count(Query):
    __slots__ = ("where_delegate",)

    def setup_delegates(self):
        self.where_delegate = WhereDelegate()

    def where(self, where: Combinable) -> Count:
        self.where_delegate.where(where)
        return self

    def response_handler(self, response) -> bool:
        return response[0]["count"]

    @property
    def querystring(self) -> QueryString:
        select = Select(self.table)
        select.where_delegate._where = self.where_delegate._where
        return QueryString(
            'SELECT COUNT(*) AS "count" FROM ({}) AS "subquery"',
            select.querystring,
        )
